//DOOM levels player by Ivaylo Beltchev - Map compiler
//email: ivob@geocities.com
//url:   http://www.geocities.com/SiliconValley/Bay/2234

#include <string.h>
#include <stdio.h>
#include <malloc.h>

#define MAXTREE 3000

struct Twall {
  double x1,y1,x2,y2;
  int    h1,h2,h3,h4;
  char   xo,yo;
  int    txw,txff,txfc,txbf,txbc;
  Twall *next;
};

struct Tbsp {
  Twall *w;
  int l,r;
};

FILE *fo;

Tbsp tree[MAXTREE];
int num;

double side(Twall *w,double x,double y)
{
  return (w->y1-w->y2)*(w->x1-x)-(w->x1-w->x2)*(w->y1-y);
}

Twall *split(Twall *w,double a,double b)
{
  Twall *w1;
  double x,y;

  w1=(Twall *)malloc(sizeof(Twall));
  x=(w->x2*a+w->x1*b)/(a+b);
  y=(w->y2*a+w->y1*b)/(a+b);
  w1->x1=x;w1->y1=y;
  w1->x2=w->x2;w1->y2=w->y2;
  w->x2=x;w->y2=y;
  w1->h1=w->h1;
  w1->h2=w->h2;
  w1->h3=w->h3;
  w1->h4=w->h4;
  w1->xo=w->xo;
  w1->yo=w->yo;
  w1->txw=w->txw;
  w1->txff=w->txff;
  w1->txfc=w->txfc;
  w1->txbf=w->txbf;
  w1->txbc=w->txbc;
  return w1;
}

Twall *getbest(Twall **l)
{
  int m=32767;
  Twall **ll,*i,*j,*w=*l;

  for (i=*l;i!=NULL;i=i->next) {
    int k=0;
    for (j=*l;j!=NULL;j=j->next)
      if (i!=j) {
        double v1=side(i,j->x1,j->y1);
        double v2=side(i,j->x2,j->y2);
        if ((v1<0 && v2>0) || (v1>0 && v2<0)) k++;
      }
    if (k<m) m=k,w=i;
  }
  for (ll=l;*ll!=NULL;ll=&((*ll)->next))
    if (*ll==w) {
      *ll=w->next;
      break;
    }
  w->next=NULL;
  return w;
}

int makebsp(Twall *l)
{
  Twall *l1,*l2,*w,*n;
  int i;

  if (l==NULL) return 0;
  i=num++;
  w=getbest(&l);
  l1=NULL;
  l2=NULL;
  for (;l!=NULL;l=n) {
    n=l->next;
    double v1=side(w,l->x1,l->y1);
    double v2=side(w,l->x2,l->y2);
    if (v1<=0 && v2<=0) {
      l->next=l1;
      l1=l;
    }
    else if (v1>=0 && v2>=0) {
      l->next=l2;
      l2=l;
    }
    else if (v1<0) {
      l->next=l1;
      l1=l;
      l=split(l,-v1,v2);
      l->next=l2;
      l2=l;
    }
    else {
      l->next=l2;
      l2=l;
      l=split(l,v1,-v2);
      l->next=l1;
      l1=l;
    }
  }
  tree[i].w=w;
  tree[i].l=makebsp(l1);
  tree[i].r=makebsp(l2);
  return i;
}

void main(int argc,char *argv[])
{
  FILE *f;
  int n,i;
  int x1,y1,x2,y2,xo,yo;
  int h1,h2,h3,h4;
  int w,ff,fc,bf,bc;
  Twall *l,**ll;
  char s[100];
  int x0,y0,z0,d0;

  printf("Map compiler for BSP.EXE by Ivaylo Beltchev ivob@geocities.com\n");
  if (argc!=2) {
    printf("Usage: makebsp <mapname>\n");
    return;
  }
  strcpy(s,argv[1]);
  if (!strchr(s,'.')) strcat(s,".MAP");
  f=fopen(s,"rt");
  if (f==NULL) {
    printf("File not found: %s\n",s);
    return;
  }
  fscanf(f,"%d %d %d %d %d",&n,&x0,&y0,&z0,&d0);
  ll=&l;
  for (i=0;i<n;i++) {
    fscanf(f,"%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
             &x1,&y1,&x2,&y2,&h1,&h2,&h3,&h4,&w,&ff,&fc,&bf,&bc,&xo,&yo);
    *ll=(Twall *)malloc(sizeof(Twall));
    (*ll)->x1=x1;
    (*ll)->y1=y1;
    (*ll)->x2=x2;
    (*ll)->y2=y2;
    (*ll)->h1=h1;
    (*ll)->h2=h2;
    (*ll)->h3=h3;
    (*ll)->h4=h4;
    (*ll)->xo=xo;
    (*ll)->yo=yo;
    (*ll)->txw=w;
    (*ll)->txff=ff;
    (*ll)->txfc=fc;
    (*ll)->txbf=bf;
    (*ll)->txbc=bc;
    (*ll)->next=NULL;
    ll=&(*ll)->next;
  }
  fclose(f);
  num=0;
  makebsp(l);
  strcpy(strchr(s,'.'),".BSP");
  fo=fopen(s,"wt");
  if (fo==NULL) {
    printf("Can't write file %s\n",s);
    return;
  }
  fprintf(fo,"%d %d %d %d %d\n",num,x0,y0,z0,d0);
  for (i=0;i<num;i++) {
    x1=tree[i].w->x1;
    y1=tree[i].w->y1;
    x2=tree[i].w->x2;
    y2=tree[i].w->y2;
    h1=tree[i].w->h1;
    h2=tree[i].w->h2;
    h3=tree[i].w->h3;
    h4=tree[i].w->h4;
    xo=tree[i].w->xo;
    yo=tree[i].w->yo;
    w =tree[i].w->txw;
    ff=tree[i].w->txff;
    fc=tree[i].w->txfc;
    bf=tree[i].w->txbf;
    bc=tree[i].w->txbc;
    fprintf(fo,"%5d %5d %5d %5d %4d %4d %4d %4d %1d %1d %1d %1d %1d %3d %3d %2d %2d\n",
      x1,y1,x2,y2,h1,h2,h3,h4,w,ff,fc,bf,bc,xo,yo,tree[i].l,tree[i].r);
  }
  fclose(fo);
}
